#include "mem_mgnt.h"
#include "RTC_EEPROM.h"
#include "globals.h"

/* ###################################################################
 * Adressen der EEPROM-Speicher
 *   0x000 - 0xFFF - I2C-EEPROM AT24C32
 *   0x000 - 0x3FF - internes EEPROM des ATmega328
 *  
 * Die Datenstruktur "SensorHour" belegt 186 (0xBA) Byte.
 * 
 * I2C-EEPROM AT24C32
 * -- SENSOR_COUNT 1 ------------------- SENSOR_COUNT 2 -------------
 * 0x000 - 0x0B9   00:00              0x000 - 0x0B9   00:00/12:00(S1)
 * 0x0C0 - 0x179   01:00              0x0C0 - 0x179   00:00/12:00(S2)
 * 0x180 - 0x239   02:00              0x180 - 0x239   01:00/13:00(S1)
 * 0x240 - 0x2F9   03:00              0x240 - 0x2F9   01:00/13:00(S2)
 * 0x300 - 0x3B9   04:00              0x300 - 0x3B9   02:00/14:00(S1)
 * 0x3C0 - 0x479   05:00              0x3C0 - 0x479   02:00/14:00(S2)
 * 0x480 - 0x539   06:00              0x480 - 0x539   03:00/15:00(S1)
 * 0x540 - 0x5F9   07:00              0x540 - 0x5F9   03:00/15:00(S2)
 * 0x600 - 0x6B9   08:00              0x600 - 0x6B9   04:00/16:00(S1)
 * 0x6C0 - 0x779   09:00              0x6C0 - 0x779   04:00/16:00(S2)
 * 0x780 - 0x839   10:00              0x780 - 0x839   05:00/17:00(S1)
 * 0x840 - 0x8F9   11:00              0x840 - 0x8F9   05:00/17:00(S2)
 * 0x900 - 0x9B9   12:00              0x900 - 0x9B9   06:00/18:00(S1)
 * 0x9C0 - 0xA79   13:00              0x9C0 - 0xA79   06:00/18:00(S2)
 * 0xA80 - 0xB39   14:00              0xA80 - 0xB39   07:00/19:00(S1)
 * 0xB40 - 0xBF9   15:00              0xB40 - 0xBF9   07:00/19:00(S2)
 * 0xC00 - 0xCB9   16:00              0xC00 - 0xCB9   08:00/20:00(S1)
 * 0xCC0 - 0xD79   17:00              0xCC0 - 0xD79   08:00/20:00(S2)
 * 0xD80 - 0xE39   18:00              0xD80 - 0xE39   09:00/21:00(S1)
 * 0xE40 - 0xEF9   19:00              0xE40 - 0xEF9   09:00/21:00(S2)
 * 
 * internes EEPROM des ATmega328
 * ------------------------------------------------------------------
 * 0x000 - 0x0B9   20:00              0x000 - 0x0B9   10:00/22:00(S1)
 * 0x0C0 - 0x179   21:00              0x0C0 - 0x179   10:00/22:00(S2)
 * 0x180 - 0x239   22:00              0x180 - 0x239   11:00/23:00(S1)
 * 0x240 - 0x2F9   23:00              0x240 - 0x2F9   11:00/23:00(S2)
 *  
 *  Die Adresse zur Stunde errechnet sich bei einem Sensor-Paar folgendermaßen:
 *    wenn Stunde>19
 *      ATmega_adr  = Blockgröße * (Stunde-20)
 *    sonst
 *      AT24C32_adr = Blockgröße * Stunde
 *
 *  Bei zwei Sensor-Paaren hingegen für das erste Sensor-Paar:
 *    wenn Stunde<12
 *      wenn Stunde>9
 *        ATmega_adr  = 2*Blockgröße * (Stunde-10)
 *      sonst
 *        AT24C32_adr = 2*Blockgröße * Stunde
 *    sonst
 *      wenn Stunde>21
 *        ATmega_adr  = 2*Blockgröße * (Stunde-22)
 *      sonst
 *        AT24C32_adr = 2*Blockgröße * (Stunde-12)
 *    für das zweite Sensor-Paar ist 1xBlockgröße zur Adresse zu addieren
 */

/* -------------------------------------------------------------------
 *  Liefert die Adresse im EEPROM für die Stunde "hour".
 *  Die Adresse ist per adr&0xFFF zu verwenden, weil 0x1000 als
 *  Kennzeichen für das interne EEPROM verwendet wird.
 */
unsigned int getEEPROMaddress(byte hour, byte sensor) {
  unsigned int adr;   // Zieladresse im EEPROM
  byte bs=sizeof(sensorHourRAM_g[0]);

  if(sensor_count_g==1) {
    if(hour>19) {
      adr=bs*(hour-20)+0x1000;                // ATmega
    } else {
      adr=bs*hour;                            // AT24C32
    }
  } else {
    if(hour<12) {
      if(hour>9) {
        adr=2*bs*(hour-10)+sensor*bs+0x1000;  // ATmega
      } else {
        adr=2*bs*hour+sensor*bs;              // AT24C32
      }
    } else {
      if(hour>21) {
        adr=2*bs*(hour-22)+sensor*bs+0x1000;  // ATmega
      } else {
        adr=2*bs*(hour-12)+sensor*bs;         // AT24C32
      }
    }   // der "sensor*bs" addiert 0*bs oder 1*bs
  }
  return(adr);
}


/* -------------------------------------------------------------------
 *  Überträgt "sensorHourRAM_g" in den EEPROM-Bereich gemäß
 *  "sensorHourRAM_g.timestamp".
 *  Über "s" wird das Sensor-Paar referenziert.
 */
void writeSensorHourToEEPROM(byte s) {
  unsigned char *ptr=(unsigned char*)&sensorHourRAM_g[s];
  unsigned int adr;        // Zieladresse im EEPROM

  // Checksum in sensorHourRAM_g[s] setzen
  sensorHourRAM_g[s].checksum=getChecksum((byte*)&sensorHourRAM_g[s], s);

  adr=getEEPROMaddress(sensorHourRAM_g[s].timestamp.hour, s);
  if(adr<0x1000) {
    writeEEPROMfull(adr, ptr, sizeof(sensorHourRAM_g[s]));
  } else {
    adr&=0xFFF;
    for(byte i=0; i<sizeof(sensorHourRAM_g[s]); i++) {
      EEPROM.write(adr++, *ptr++);
    }
  }
}


/* -------------------------------------------------------------------
 *  Kopiert den EEPROM-Inhalt für die Stunde "hour" und das
 *  Sensor-Paar "sensor" nach "buf". "buf" muss mindestens 186 Byte
 *  Platz haben.
 *  Liefert True, wenn die Checksum gepasst hat.
 */
char readHourFromEEPROM(byte hour, byte sensor, byte *buf) {
  unsigned int adr;

  adr=getEEPROMaddress(hour, sensor);
  //Serial.print("EEPROM:"); Serial.println(adr, HEX);
  if(adr<0x1000) {
    readEEPROMfull(adr, buf, 186);
  } else {
    adr&=0xFFF;
    for(int i=0; i<186; i++) {
      buf[i]=EEPROM.read(adr++);
    }
  }
  adr=buf[184]+buf[185]*256;   // Checksum rauslösen
  return(adr==getChecksum(buf, sensor));
}

/* -------------------------------------------------------------------
 *  Liefert die Checksum von "ptr.timestamp" und "ptr.data[60]".
 */
unsigned int getChecksum(byte *ptr, byte s) {
  unsigned int cs;
  cs=0;
  for(byte i=0; i<184; i++) {   // über 1xRTCdata + 60xSensorData
    cs+=*ptr++;
    cs++;
  }
  return(cs);
}


/* -------------------------------------------------------------------
 *  Setzt alle SensorHour-Blöcke in beiden EEPROMs auf ein ungültiges
 *  RTCdata.
 */
void invalidateSensorHours(void) {
  byte i;
  int adr;
  unsigned long invalid=0xffffffff;   // ungültiges RTCdata

  digitalWrite(RTC_EEPROM_POWER_PIN, HIGH);
  for(adr=0; adr<=0xe40; adr+=0xc0) {
    writeEEPROM(adr, (unsigned char*)&invalid, 4);
  }
  digitalWrite(RTC_EEPROM_POWER_PIN, LOW);
  for(adr=0; adr<=0x240; adr+=0xc0) {
    for(i=0; i<4; i++) {
      EEPROM.write(adr, 0xff);
    }
  }
}


/* -------------------------------------------------------------------
 *  Debug-Ausgabe
 */
void dumpATmegaEEPROM(void) {
  unsigned char buf[16];
  char c[7], j;

  for(int i=0; i<1024; i+=16) {         // über das gesamte EEPROM
    snprintf(c, 7, "%04x  ", i);        // Adresse ausgeben
    Serial.print(c);
    for(int k=0; k<16; k++) {           // 16 Byte lesen
      buf[k]=EEPROM.read(i+k);
    }
    for(j=0; j<16; j++) {
      snprintf(c, 4, " %02x", buf[j]);  // 16 HEX-Werte ausgeben
      Serial.print(c);
    }
    Serial.print("  ");
    for(j=0; j<16; j++) {               // 16 ASCII-Werte
      if(buf[j]>=32 && buf[j]<128) {    // nur darstellbare Zeichen...
        snprintf(c, 2, "%c", buf[j]);
        Serial.print(c);                // ...ausgeben
      } else {
        Serial.print(".");              // sonst stattdessen "." ausgeben
      }
    }
    Serial.println();
  }
}

